#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <setjmp.h>
#include <unistd.h>
#include <signal.h>
#include <readline/readline.h>
#include <readline/history.h>
#include "lac.h"

lenv_t *null_env;

void
sigint (int sig)
{
  raise_exception ("Interrupted", NIL);
}

#define TDIFF_SEC(_a, _b)				\
    ((_b)->tv_usec >= (_a)->tv_usec			\
     ? (_b)->tv_sec - (_a)->tv_sec			\
     : (_b)->tv_sec - (_a)->tv_sec - 1)

#define TDIFF_USEC(_a, _b)						\
    ((_b)->tv_usec >= t1.tv_usec					\
     ? (_b)->tv_usec - (_a)->tv_usec					\
     : (_b)->tv_usec + 1000000L - (_a)->tv_usec)

lreg_t
timed_eval (lreg_t lr, void *opq)
{
  lreg_t ret;
  struct timeval t1, t2;
  lenv_t *env = (lenv_t *) opq;

  gettimeofday (&t1, NULL);
  ret = lac_eval (lr, env);
  gettimeofday (&t2, NULL);

  fprintf (stderr, "Evaluation took %ld seconds and %ld microseconds.\n",
	   TDIFF_SEC (&t1, &t2), TDIFF_USEC (&t1, &t2));

  return ret;
}

int
repl (FILE * infd, FILE * outfd, FILE * errfd)
{
  int r = 0;
  char *buf = NULL;
  lreg_t res = NIL;

restart:
  lac_on_error ({
		 lac_error_print (errfd);
		 res = NIL;
		 goto restart;
    });

  do
    {
      buf = readline ("LAC>");
      if (buf == NULL)
	{
	  r = -1;
	  break;
	}
      if (*buf != '\0')
	add_history (buf);

      res = sexpr_parse_string (buf, timed_eval, (void *) null_env);
      if (isatty (fileno (outfd)))
	{
	  fprintf (outfd, "=> ");
	  sexpr_fprint (outfd, res);
	  fprintf (outfd, "\n");
	}
    }
  while (1);

  lac_off_error ();
  return r;
}

int
main (int argc, char *argv[])
{

  signal (SIGINT, sigint);

  null_env = lac_init ();
  if (null_env == NULL)
    {
      fprintf (stderr, "Initialization failed.\n");
      return -1;
    }
  repl (stdin, stdout, stderr);
  fprintf (stdout, "\ngoodbye!\n");
  return 0;
}
